#include <assert.h>

#include <math.h>
#include <signal.h>
#include <unistd.h>
#include <algorithm>
#include <array>
#include <atomic>
#include <chrono>
#include <fstream>
#include <iomanip>
#include <iostream>
#include <memory>
#include <mutex>
#include <queue>
#include <string>
#include <thread>
#include <vector>
#include <opencv2/core/core.hpp>
#include <opencv2/highgui/highgui.hpp>
#include <opencv2/imgproc/imgproc.hpp>
#include <opencv2/opencv.hpp>

// Header files for DNNDK API
#include <dnndk/dnndk.h>
#include "dputils.h"

// Un-comment following define for application without any I/O
//#define WITHOUT_IO

// user configurable settings
// Default to TensorFlow network model, without verbose
int bCaffe = 0;
int bTensorFlow = 1;
int bDisplayBinary = 0;
int bVerbose = 0;

/* DPU Kernel, Input Node, Output Node names for TensorFlow version of network */
#define KERNEL_TF "customcnn"
#define INPUT_NODE_TF "conv2d_Conv2D"
#define OUTPUT_NODE_TF "conv2d_3_Conv2D"

/* DPU Kernel, Input Node, Output Node names for Caffe version of network */
#define KERNEL_CAFFE "LeNet_0"
#define INPUT_NODE_CAFFE "conv1"
#define OUTPUT_NODE_CAFFE "fc2"


using namespace std;
using namespace std::chrono;
using namespace cv;

typedef pair<int, Mat> pairImage;

class PairComp {  // An auxiliary class for sort the image pair according to its
                  // index
    public:
    bool operator()(const pairImage &n1, const pairImage &n2) const {
        if (n1.first == n2.first) return n1.first > n2.first;
        return n1.first > n2.first;
    }
};

/**
 * @brief Set image into DPU Task for MNIST
 *
 * @param taskMNIST - pointer to MNIST Task
 * @param input_node - name of input layer
 * @param image - input image
 *
 * @return none
 */
void setInputImageMNIST(DPUTask *task, const char* input_node, const cv::Mat& image){
	DPUTensor* dpu_in = dpuGetInputTensor(task, input_node);
	float scale = dpuGetTensorScale(dpu_in);
	int width = dpuGetTensorWidth(dpu_in);
	int height = dpuGetTensorHeight(dpu_in);
	if ( width != image.cols || height != image.rows )
	{
		cout << "ERROR: Incorrect input image size " << image.cols << "x" << image.rows << " !" << endl;
		exit(0);
	}
	int size = dpuGetTensorSize(dpu_in);
	int8_t* data = dpuGetTensorAddress(dpu_in);

	for(int i = 0; i < size; ++i) {
		//data[i] = (float(image.data[i])/255.0 - 0.5)*2 * scale;
		data[i] = (float(image.data[i])/255.0 ) * scale;
	}
}

/**
 * @brief Get top k results according to its probability
 *
 * @param d - pointer to input data
 * @param size - size of input data
 * @param k - calculation result
 *
 * @return none
 */
void TopK(const float *d, int size, int k) {
    assert(d && size > 0 && k > 0);
    priority_queue<pair<float, int>> q;

    for (auto i = 0; i < size; ++i) {
        q.push(pair<float, int>(d[i], i));
    }

    for (auto i = 0; i < k; ++i) {
        pair<float, int> ki = q.top();
        printf("top[%d] prob = %-8f digit = %d\n", i, d[ki.second], ki.second );
        q.pop();
    }
}

/**
 * @brief runMNIST - Run DPU Task for MNIST
 *
 * @param task - pointer to a DPU Task
 * @param img  - input image in OpenCV's Mat format
 *
 * @return none
 */
void runMNIST(DPUTask *taskMNIST, Mat &img) {

    /* Get channel count of the output Tensor for MNIST Task  */
    int channel = dpuGetOutputTensorChannel(taskMNIST, bTensorFlow?OUTPUT_NODE_TF:OUTPUT_NODE_CAFFE);
    // Get the scale of classification result
    float scale = dpuGetOutputTensorScale(taskMNIST, bTensorFlow?OUTPUT_NODE_TF:OUTPUT_NODE_CAFFE);

    int8_t *DPUResult = new int8_t[channel];
    float *softmax = new float[channel];

    Mat step1;
	Mat step2;

	vector<vector<Point>> contours;
	vector<Vec4i> hierarchy;
	vector<Rect> boxes;

	cvtColor(img, step1, COLOR_RGB2GRAY);
	//GaussianBlur(step1, step1, Size(5, 5), 0, 0);
	threshold(step1, step1, 127, 255, THRESH_BINARY_INV);
	step1.copyTo(step2);
	findContours(step2, contours, hierarchy, CV_RETR_EXTERNAL, CHAIN_APPROX_SIMPLE);

	if ( bDisplayBinary )
	{
		cvtColor(step1,img, COLOR_GRAY2RGB);
	}

	for (int i = 0; i < (int)contours.size(); i++)
	{
		Rect bounding = boundingRect(contours[i]);
		int cx = bounding.x + bounding.width / 2;
		int cy = bounding.y + bounding.height / 2;

		bounding.width = bounding.height = max(bounding.width, bounding.height) * 1.3;
		bounding.x = cx - bounding.width / 2;
		bounding.y = cy - bounding.height / 2;

		if (bounding.x < 0) continue;
		if (bounding.y < 0) continue;
		if (bounding.x + bounding.width  > 640) continue;
		if (bounding.y + bounding.height > 480) continue;
		if (bounding.area() < 2500) continue;

		//printf("bounding: %i, %i, %i, %i\r\n", bounding.x, bounding.y, bounding.width, bounding.height);
		rectangle(img, bounding, Scalar(0, 255, 0), 3);

		Mat digit = step1(bounding);

		//imwrite("digit.jpg", digit);
		resize(digit, digit, Size(28, 28));

	    /* Set image into DPU Task for MNIST */
	    if ( bTensorFlow )
	    {
	    	setInputImageMNIST( taskMNIST, INPUT_NODE_TF, digit );
	    }
	    else
	    {
			// convert back to RGB for neural network
			Mat digit_rgb;
			cvtColor(digit, digit_rgb, COLOR_GRAY2RGB);

			dpuSetInputImage2( taskMNIST, INPUT_NODE_CAFFE, digit_rgb );
	    }

	    /* Launch MNIST Task */
	    dpuRunTask(taskMNIST);

	    // Get the output tensor from DPU in DPU INT8 format
 //       DPUTensor* dpuOutTensorInt8 = dpuGetOutputTensorInHWCInt8(taskMNIST, bTensorFlow?OUTPUT_NODE_TF:OUTPUT_NODE_CAFFE);
         dpuGetOutputTensorInHWCInt8(taskMNIST, bTensorFlow?OUTPUT_NODE_TF:OUTPUT_NODE_CAFFE,DPUResult,channel);
		// Get the data pointer from the output tensor
//	    DPUResult = dpuGetTensorAddress(dpuOutTensorInt8);
	    // Processing softmax in DPU with batchsize=1
	    dpuRunSoftmax( DPUResult, softmax, channel, 1, scale);

	    if ( bVerbose )
	    {
		    /* Display TOP-5 classification results */
	    	TopK(softmax, channel, 5);
	    }

	    unsigned result = 0;
	    float score = 0.0;
	    for ( int i = 0; i < channel; i++ ) {
	    	if ( softmax[i] > score ) {
	    		result = i;
	    		score = softmax[i];
	    	}
	    }

	    if ( bVerbose )
	    {
	    	cout << "Digit = " << result << endl;
	    }

		String text = to_string(result);
		putText(img, text, bounding.tl(), FONT_HERSHEY_SCRIPT_SIMPLEX, 2, Scalar(0, 0, 255), 2);

	}

}

/*
 * @brief digitRecognition - Entry of digit recognition using MNIST
 *
 * @param kernel - point to DPU Kernel
 *
 * @return none
 */
void digitRecognition(DPUKernel *kernel) {
    mutex mtxQueueInput;                                               // mutex of input queue
    mutex mtxQueueShow;                                                // mutex of display queue
    queue<pairImage> queueInput;                                       // input queue
    priority_queue<pairImage, vector<pairImage>, PairComp> queueShow;  // display queue

    int idxInputImage = 0;
    int dpuFrameCount = 0;
    int displayFrameCount = 0;

#if !defined(WITHOUT_IO)
    VideoCapture camera(0);
    if (!camera.isOpened()) {
        cerr << "Open camera error!" << endl;
        exit(-1);
    }
#endif

    // We create three different threads to do face detection:
    // 1. Reader thread  : Read images from camera and put it to the input queue;
    //
    // 2. Worker threads : Each worker thread repeats the following 3 steps util
    // no images:
    // (1) get an image from input queue;
    // (2) process it using DenseBox model;
    // (3) put the processed image to the display queue.
    //
    // 3. Display thread : Get output image from queueShow and display it.

    // 1. Reader thread
    atomic<bool> bReading(true);
    thread reader([&]() {
        // image index of input video
        while (true) {
#if !defined(WITHOUT_IO)
            Mat img;
            camera >> img;
            if (img.empty()) {
                cerr << "Fail to read image from camera!" << endl;
                camera.release();
                break;
            }
#else
            // Create blank black image
            Mat img(480,640,CV_8UC3,Scalar(0,0,0));
            usleep(33000); // 33msec for 30fps
#endif
            mtxQueueInput.lock();
            queueInput.push(make_pair(idxInputImage++, img));
            if (queueInput.size() >= 100) {
                mtxQueueInput.unlock();
                cout << "[Warning]input queue size is " << queueInput.size() << endl;
                // Sleep for a moment
                sleep(2);
            } else {
                mtxQueueInput.unlock();
            }
        }
        bReading.store(false);
    });

    // 2. Worker thread
    constexpr int workerNum = 2;
    thread workers[workerNum];
    atomic<int> workerAlive(workerNum);
    for (auto i = 0; i < workerNum; i++) {
        workers[i] = thread([&]() {
            // Create DPU Tasks from DPU Kernel
            DPUTask *task = dpuCreateTask(kernel, 0);

            while (true) {
                pair<int, Mat> pairIndexImage;
                mtxQueueInput.lock();
                if (queueInput.empty()) {
                    mtxQueueInput.unlock();
                    if (bReading.load())
                        continue;
                    else
                        break;
                } else {
                    // Get an image from input queue
                    pairIndexImage = queueInput.front();
                    queueInput.pop();
                }
                mtxQueueInput.unlock();
                // Process the image using MNIST model
                runMNIST(task, pairIndexImage.second);
                mtxQueueShow.lock();
                // Put the processed image to show queue
                queueShow.push(pairIndexImage);
                dpuFrameCount++;
                mtxQueueShow.unlock();
            }

            // Destroy DPU Tasks & free resources
            dpuDestroyTask(task);

            workerAlive--;
        });
    }

    // 3. Display thread;
    atomic<int> idxShowImage(0);  // next frame index to be display
    thread show([&]() {
        while (true) {
            mtxQueueShow.lock();
            if (queueShow.empty()) {  // no image in display queue
                mtxQueueShow.unlock();
                if (workerAlive.load() == 0) {
                    cout << "MNIST End." << endl;
                    break;
                } else {
                    usleep(10000);  // Sleep for a moment
                }
            } else if (idxShowImage.load() == queueShow.top().first) {
#if 1
            	while ( queueShow.size() > 2 ) {
                    idxShowImage++;
                    queueShow.pop();
            	}
#endif
#if !defined(WITHOUT_IO)
                cv::imshow("MNIST",
                           queueShow.top().second);  // Display image
#endif
                displayFrameCount++;
                idxShowImage++;
                queueShow.pop();
                mtxQueueShow.unlock();
                if (waitKey(1) == 'q') {
                    bReading = false;
                    exit(0);
                }

            } else {
                mtxQueueShow.unlock();
            }
        }
    });

    // 4. Stats thread;
    thread stats([&]() {
        int idxInputImage_prev = 0;
        int dpuFrameCount_prev = 0;
        int displayFrameCount_prev = 0;

        while (true) {
            printf("Input frames: %d (%d), DPU Frames: %d (%d), Display Frames: %d (%d)\n",
            		idxInputImage, (idxInputImage - idxInputImage_prev),
					dpuFrameCount, (dpuFrameCount - dpuFrameCount_prev),
					displayFrameCount, (displayFrameCount - displayFrameCount_prev)
					);
            idxInputImage_prev = idxInputImage;
            dpuFrameCount_prev = dpuFrameCount;
            displayFrameCount_prev = displayFrameCount;
            usleep(1000000);
        }
    });

    // Release thread resources.
    if (reader.joinable()) reader.join();
    if (show.joinable()) show.join();
    for (auto &w : workers) {
        if (w.joinable()) w.join();
    }
}

static void usage (char *argv0)
{
	char *basename = strrchr(argv0, '/');
	if (!basename)
		basename = argv0;

	fprintf(stderr,
		"Usage: %s [-c] [-t] [-v]\n"
		"\t\t-c\t\tUse Caffe version of MNIST network model (LeNet)\n"
		"\t\t-t\t\tUse TensorFlow version of MNIST network model (moist)\n"
		"\t\t-b\t\tDisplay binary image\n"
		"\t\t-v\t\tVerbose\n"
		"\n"
		,  basename);
	exit(-2);
}

/*
 * @brief main - Entry of MNIST live I/O application.
 *
 * @note This sample presents how to implement
 *       an MNIST digit classification application on DPU
 *       using MNIST (TensorFlow) or LENET (Caffe) network models.
 *
 */
int main(int argc, char *argv[])
{
	int c;

	while ((c = getopt(argc, argv, "ctbv")) != -1) {
		switch (c) {
			case 'c':
				printf( "Using Caffe version of network model (LeNet)\r\n" );
				bCaffe = 1;
				bTensorFlow = 0;
				break;
			case 't':
				printf( "Using TensorFlow version of network model (mnist)\r\n" );
				bCaffe = 0;
				bTensorFlow = 1;
				break;
			case 'b':
				printf( "Display binary image\r\n" );
				bDisplayBinary = 1;
				break;
			case 'v':
				printf( "Enabling verbose\r\n" );
				bVerbose = 1;
				break;
			case '?':
				usage(argv[0]);
			default:
				usage(argv[0]);

		}
	}

    // Attach to DPU driver and prepare for running
    dpuOpen();

    // Load DPU Kernel for MNIST neural network
    DPUKernel *kernel;
    if ( bTensorFlow )
    	kernel = dpuLoadKernel(KERNEL_TF);
    else
    	kernel = dpuLoadKernel(KERNEL_CAFFE);

    // Doing digit recognition.
    digitRecognition(kernel);

    // Destroy DPU Kernel & free resources
    dpuDestroyKernel(kernel);

    // Detach from DPU driver & release resources
    dpuClose();

    return 0;
}
